/* $Id: fflow.c,v 1.15 2002/11/05 10:15:24 aa5779 Exp $ */
/*
	Copyright (C) 2001, 2002 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "dstring.h"
#include "flow.h"
#define FFLOW_IMP
#include "fflow.h"
#include "util.h"

/// \subsection[FILE FLOW]
// Files flows have 'file://' prefix in their names (if no prefix is used,
// file:// is assumed). The rest of the name is treated as a file name.
// If it is empty, stdin/stdout is assumed.
// 
// If a flow is used for reading the file must exist; if a flow is opened
// for writing, it is created or truncated if it already exists.
// It is not possible to use the same flow for reading and writing.
//
// If a flow is readable and it is not connected to stdin, it tries to mmap
// a file; if mapping fails, the whole contents of the file is read into
// memory -- you should keep it in mind when dealing with large files.


flow_class file_flow_class = {"file", ff_open, ff_close, ff_getc, ff_ungetc, 
							   ff_putc, ff_puts,
							   ff_getpos, ff_setpos, ff_rewind,
							   ff_flowctl
							 };
flow_class *default_flow_class = &file_flow_class;

// \paragraph[options recognized by file flows]
// \item \em[exist] do not create/truncate a file when opening for writing;
// the file \em[must] exist
// \item \em[append] open a file for appending
// \item \em[noload] do not map/load a file into memory (useful for special devices etc)
flow_handle_t ff_open(char *name, char *options, int writeable)
{
	file_flow_data *ffd = ALLOC(file_flow_data);
	ffd->writeable = writeable;
	if(writeable)
	{
		ffd->preload = 0;
		if(name && *name)
			ffd->x.hfile = xfopen(name, strstr(options, "append") ? "a" : 
											(strstr(options, "exist") ? "r+" : "w"));
		else ffd->x.hfile = stdout;
	}
	else
	{	
		if(name && *name)
		{
			if(!strstr(options, "noload"))
			{
				ffd->x.inmap = map_file(name, NULL, NULL);
				ffd->pos = ffd->x.inmap->data;
				ffd->preload = 1;
			}
			else ffd->x.hfile = xfopen(name, "r");
		}
		else ffd->x.hfile = stdin;
	}
	ffd->undone = FLOW_EOF;
	ffd->buf_size = BUFSIZ;
	return ffd;
}

void ff_close(flow_handle_t fh)
{
	file_flow_data *ffd = fh;
	if(!ffd->preload)
	{
		if(ffd->x.hfile != stdin && ffd->x.hfile != stdout)
			fclose(ffd->x.hfile);
	}
	else delete_mapping(ffd->x.inmap);
	FREE(ffd);
}

int ff_getc(flow_handle_t fh)
{
	file_flow_data *ffd = fh;
	if(ffd->undone != FLOW_EOF)
	{
		int ch = ffd->undone;
		ffd->undone = FLOW_EOF;
		return ch;
	}
	if(ffd->preload)
		return is_map_eof(ffd->x.inmap, ffd->pos) ? FLOW_EOF : *ffd->pos++;
	else
	{
		int ch;
		
		ch = fgetc(ffd->x.hfile);
	    if(ch != EOF)
			return ch;
		if(ferror(ffd->x.hfile))
			return FLOW_ERROR;
		return FLOW_EOF;
	}
}

int ff_ungetc(flow_handle_t fh, int ch)
{
	file_flow_data *ffd = fh;

	if(ffd->undone != FLOW_EOF)
	{
		errno = EBUSY;
		return FLOW_ERROR;
	}
	ffd->undone = ch;
	return 0;
}

int ff_putc(flow_handle_t fh, int ch)
{
	file_flow_data *ffd = fh;
	
	if(!ffd->writeable)
	{
		errno = EACCES;
		return FLOW_ERROR;
	}
	return fputc(ch, ffd->x.hfile) == EOF ? FLOW_ERROR : 0;
}

int ff_puts(flow_handle_t fh, char *str)
{
	file_flow_data *ffd = fh;
	if(!str)
	{
		errno = EINVAL;
		return FLOW_ERROR;
	}
	if(!ffd->writeable)
	{
		errno = EACCES;
		return FLOW_ERROR;
	}
	return fputs(str, ffd->x.hfile) == EOF ? FLOW_ERROR : 0; 
}
 
flow_pos_t ff_getpos(flow_handle_t fh)
{
	file_flow_data *ffd = fh;
	int pos;
	int correct = ffd->undone != FLOW_EOF;

	if(ffd->preload)
		return ffd->pos - (char *)ffd->x.inmap->data - correct;
	pos = ftell(ffd->x.hfile);
	return pos == -1 ? FLOW_ERROR : pos - correct;
}

int ff_setpos(flow_handle_t fh, flow_pos_t fp)
{
	file_flow_data *ffd = fh;
	ffd->undone = FLOW_EOF;
	if(ffd->preload)
	{
		if(fp < 0 || fp >= ffd->x.inmap->len)
		{
			errno = EINVAL;
			return FLOW_ERROR;
		}
		ffd->pos = ffd->x.inmap->data + fp;
		return 0;
	}
	return fseek(ffd->x.hfile, SEEK_SET, fp) == -1 ? FLOW_ERROR : 0;
}

int ff_rewind(flow_handle_t fh)
{
	file_flow_data *ffd = fh;
	ffd->undone = FLOW_EOF; 
	if(ffd->preload)
	{
		ffd->pos = ffd->x.inmap->data;
		return 0;
	}
	rewind(ffd->x.hfile);
	return 0;
}

static char *change_buffering(file_flow_data *ffd, int mode)
{
	if(!ffd->preload)
	{
		if(ffd->writeable)
			fflush(ffd->x.hfile);
		setvbuf(ffd->x.hfile, NULL, mode, ffd->buf_size);
	}
	return "";
}

// \paragraph[ioctls recognized by file flows]
// \item \em[isatty] -- yields true when a flow is connected to a terminal
// \item \em[loaded?] -- returns true if a file has been loaded into memory
// \item \em[flush] -- flushes the output; no effect on input
// \item \em2[unbuffer][,] \em2[lbuffer][,] \em[buffer] -- changes the
// buffering mode of a flow to none, line-oriented or full
// \item [bufsize=n] -- sets the size of the I/O buffer for an associated file to \em[n].
// If \em[n] is 0, a system default is assumed.
// 
// The setting to take an effect, you must then issue \em[buffer] or \em[lbuffer]. 
// \item \em[fileno] -- gets an OS file handle associated with a flow (an integer);
// empty string if it is unavailable
// \item \em[lock] -- tries to lock a file associated with a flow; returns an
// empty string if successful (or the system doesn't support locking at all);
// \em[unable] in case of a failure (usually due to a lock held by another process)
//
// The file is locked for writing if the flow is writeable; for reading otherwise.
// \item \em[unlock] -- tries to unlock an associated file
// \item \em[lock?] -- returns an empty string if there is no locks on the file;
// otherwise a string in the form \em[type start len pid] where \em[type] is
// either \em[read] or \em[write]; \em[start] is a starting point of a locked region
// (0 for our own locks); \em[len] is the length of a locked region (0 for our own
// locks); \em[pid] is the lock owner's pid
// \endlist
char *ff_flowctl(flow_handle_t fh, char *ctl)
{
	file_flow_data *ffd = fh;
#ifdef HAVE_ISATTY	
#ifdef HAVE_FILENO
	if(!strcmp(ctl, "isatty"))
		return ffd->preload ? "" : (isatty(fileno(ffd->x.hfile)) ? "true" : "");
	else 
#endif		
#endif		
	if(!strcmp(ctl, "loaded?"))
		return ffd->preload ? "true" : "";
	else if(!strcmp(ctl, "flush"))
	{
		if(ffd->writeable)
			fflush(ffd->x.hfile);
		return "";
	}
	else if(!strcmp(ctl, "unbuffer"))
		return change_buffering(ffd, _IONBF);
	else if(!strcmp(ctl, "lbuffer"))
		return change_buffering(ffd, _IOLBF);
	else if(!strcmp(ctl, "buffer"))
		return change_buffering(ffd, _IOFBF);
	else if(!strncmp(ctl, "bufsize=", 8))
	{
		ffd->buf_size = strtoul(ctl + 8, NULL, 0);
		if(!ffd->buf_size)
			ffd->buf_size = BUFSIZ;
		return "";
	}
	else if(!strcmp(ctl, "bufsize?"))
		return DS_BODY(ds_fromint(ffd->buf_size, 10, 0));
#if defined(CAN_LOCK) && defined(HAVE_FILENO)
	else if(!strcmp(ctl, "lock"))
	{
		struct flock lock;
		lock.l_type  = !ffd->writeable ? F_RDLCK : F_WRLCK;
		lock.l_start = 0;
		lock.l_whence = SEEK_SET;
		lock.l_len	 = 0;
		return fcntl(ffd->preload ? ffd->x.inmap->hfile : fileno(ffd->x.hfile), F_SETLK, &lock) ? "unable" : "";
	}
	else if(!strcmp(ctl, "unlock"))
	{
		struct flock lock;
		lock.l_type = F_UNLCK;
		lock.l_start = 0;
		lock.l_whence = SEEK_SET;
		lock.l_len = 0;
		return fcntl(ffd->preload ? ffd->x.inmap->hfile : fileno(ffd->x.hfile), F_SETLK, &lock) ? "unable" : "";
	}
	else if(!strcmp(ctl, "lock?"))
	{
		struct flock lock;
		static char buffer[128];
		lock.l_type = ffd->writeable ? F_WRLCK : F_RDLCK;
		lock.l_start  = 0;
		lock.l_whence = SEEK_SET;
		lock.l_len    = 0;
		if(fcntl(ffd->preload ? ffd->x.inmap->hfile : fileno(ffd->x.hfile), F_GETLK, &lock))
			recover_error("can't get lock: %s", strerror(errno));
		if(lock.l_type == F_UNLCK)
			return "";
		sprintf(buffer, "%s %u %u %u", lock.l_type == F_RDLCK ? "read" : "write",
						(unsigned)lock.l_start, (unsigned)lock.l_len, lock.l_pid);
		return buffer;
	}
#else
	else if(!strcmp(ctl, "lock") || !strcmp(ctl, "unlock") || !strcmp(ctl, "lock?"))
		return "";
#endif	
#ifdef HAVE_FILENO
	else if(!strcmp(ctl, "fileno"))
	{
		return DS_BODY(ds_fromint(ffd->preload ? ffd->x.inmap->hfile : fileno(ffd->x.hfile), 10, 0));
	}
#else
	else if(!strcmp(ctl, "fileno"))
		return "";
#endif	
	return NULL;
}

